__author__    = 'Jonathan Yang <yuekyang@cisco.com>'
__date__      = 'July 22 2016'
__contact__   = 'yang-python@cisco.com'
__copyright__ = 'Cisco System, Inc. Cisco Confedential'

__import__('pkg_resources').declare_namespace(__name__)
